from httplib import HTTPConnection, HTTPSConnection
from urlparse import urlparse
from warnings import warn
import base64
import sys

from mustaine.encoder import encode_object
from mustaine.parser import Parser
from mustaine.protocol import Call, Fault
from mustaine._util import BufferedReader
from mustaine import __version__


class ProtocolError(Exception):
    """ Raised when an HTTP error occurs """
    def __init__(self, url, status, reason):
        self._url    = url
        self._status = status
        self._reason = reason

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<ProtocolError for %s: %s %s>" % (self._url, self._status, self._reason,)


class HessianProxy(object):

    def __init__(self, service_uri, credentials=None, key_file=None, cert_file=None, timeout=10, buffer_size=65535, error_factory=lambda x: x):
        self._headers = list()
        self._headers.append(('User-Agent', 'mustaine/' + __version__,))
        self._headers.append(('Content-Type', 'application/x-hessian',))

        if sys.version_info < (2,6):
            warn('HessianProxy timeout not enforceable before Python 2.6', RuntimeWarning, stacklevel=2)
            timeout = {}
        else:
            timeout = {'timeout': timeout}

        self._uri = urlparse(service_uri)
        if self._uri.scheme == 'http':
            self._client = HTTPConnection(self._uri.hostname,
                                          self._uri.port or 80,
                                          strict=True,
                                          **timeout)
        elif self._uri.scheme == 'https':
            self._client = HTTPSConnection(self._uri.hostname,
                                           self._uri.port or 443,
                                           key_file=key_file,
                                           cert_file=cert_file,
                                           strict=True,
                                           **timeout)
        else:
            raise NotImplementedError("HessianProxy only supports http:// and https:// URIs")

        # autofill credentials if they were passed via url instead of kwargs
        if (self._uri.username and self._uri.password) and not credentials:
            credentials = (self._uri.username, self._uri.password)

        if credentials:
            auth = 'Basic ' + base64.b64encode(':'.join(credentials))
            self._headers.append(('Authorization', auth))

        self._buffer_size = buffer_size
        self._error_factory = error_factory
        self._parser = Parser()

    class __RemoteMethod(object):
        # dark magic for autoloading methods
        def __init__(self, caller, method):
            self.__caller = caller
            self.__method = method
        def __call__(self, *args):
            return self.__caller(self.__method, args)

    def __getattr__(self, method):
        return self.__RemoteMethod(self, method)

    def __repr__(self):
        return "<mustaine.client.HessianProxy(\"%s\")>" % (self._uri.geturl(),)

    def __str__(self):
        return self.__repr__()

    def __call__(self, method, args):
        try:
            self._client.putrequest('POST', self._uri.path)
            for header in self._headers:
                self._client.putheader(*header)

            request = encode_object(Call(method, args))
            self._client.putheader("Content-Length", str(len(request)))
            self._client.endheaders()
            self._client.send(str(request))

            response = self._client.getresponse()
            if response.status != 200:
                raise ProtocolError(self._uri.geturl(), response.status, response.reason)

            length = response.getheader('Content-Length', -1)
            if length == '0':
                raise ProtocolError(self._uri.geturl(), 'FATAL:', 'Server sent zero-length response')

            reply = self._parser.parse_stream(BufferedReader(response, buffer_size=self._buffer_size))
            self._client.close()

            if isinstance(reply.value, Fault):
                raise self._error_factory(reply.value)
            else:
                return reply.value
        except:
            self._client.close()
            raise



try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

class BufferedReader(object):

    def __init__(self, input, buffer_size=65535):
        self.__input       = input
        self.__buffer_size = buffer_size
        self.__buffer      = StringIO()

        # initial fill
        chunk = input.read(buffer_size)
        self.__byte_count = len(chunk)
        self.__buffer.write(chunk)
        self.__buffer.seek(0)

    def read(self, byte_count):
        difference = byte_count - self.__byte_count

        if difference < 0:
            chunk = self.__buffer.read(byte_count)
            self.__byte_count -= byte_count
        else:
            chunk = self.__buffer.read() + self.__input.read(difference)

            # verify size
            if len(chunk) != byte_count:
                raise EOFError("Encountered unexpected end of stream")

            # reset internal buffer
            self.__buffer.seek(0)
            self.__buffer.truncate()

            # replenish
            fresh_chunk = self.__input.read(self.__buffer_size)
            self.__byte_count = len(fresh_chunk)
            self.__buffer.write(fresh_chunk)
            self.__buffer.seek(0)

        return chunk




# transparent types used for hessian serialization
# objects of this type can appear on the wire but have no native python type

class Call(object):
    def __init__(self, method=None, args=None, headers=None, overload=None):
        self._method   = method or ''
        self._args     = args or list()
        self._headers  = headers or dict()
        self._overload = overload or False

    def _get_method(self):
        return self._method

    def _set_method(self, value):
        if isinstance(value, str):
            self._method = value
        else:
            raise TypeError("Call.method must be a string")

    method = property(_get_method, _set_method)

    def _get_args(self):
        return self._args

    def _set_args(self, value):
        if hasattr(value, '__iter__'):
            self._args = value
        else:
            raise TypeError("Call.args must be an iterable value")

    args = property(_get_args, _set_args)

    def _get_headers(self):
        return self._headers

    def _set_headers(self, value):
        if not isinstance(value, dict):
            raise TypeError("Call.headers must be a dict of strings to objects")

        for key in value.keys():
            if not isinstance(key, basestring):
                raise TypeError("Call.headers must be a dict of strings to objects")

        self._headers = value

    headers = property(_get_headers, _set_headers)

    def _get_overload(self):
        return self._overload

    def _set_overload(self, value):
        if isinstance(value, bool):
            self._overload = value
        else:
            raise TypeError("Call.overload must be True or False")

    overload = property(_get_overload, _set_overload)


class Reply(object):
    def __init__(self, value=None, headers=None):
        self.value    = value # unmanaged property
        self._headers = headers or dict()

    def _get_headers(self):
        return self._headers

    def _set_headers(self, value):
        if not isinstance(value, dict):
            raise TypeError("Call.headers must be a dict of strings to objects")

        for key in value.keys():
            if not isinstance(key, basestring):
                raise TypeError("Call.headers must be a dict of strings to objects")

        self._headers = value

    headers = property(_get_headers, _set_headers)


class Fault(Exception):
    def __init__(self, code, message, detail):
        self.code    = code
        self.message = message 
        self.detail  = detail

    # 'message' property implemented to mask DeprecationWarning
    def _get_message(self):
        return self.__message

    def _set_message(self, message):
        self.__message = message

    message = property(_get_message, _set_message)

    def __repr__(self):
        return "<mustaine.protocol.Fault: \"%s: %s\">" % (self.code, self.message,)

    def __str__(self):
        return self.__repr__()


class Binary(object):
    def __init__(self, value):
        self.value = value


class Remote(object):
    def __init__(self, type_name=None, url=None):
        self.type_name = type_name
        self.url       = url


class Object(object):
    def __init__(self, meta_type, **kwargs):
        self.__meta_type = meta_type

        if kwargs:
            for key in kwargs:
                self.__dict__[key] = kwargs[key]

    @property
    def _meta_type(self):
        return self.__meta_type

    def __repr__(self):
        return "<%s object at %s>" % (self.__meta_type, hex(id(self)),)

    def __getstate__(self):
        # clear metadata for clean pickling
        t = self.__meta_type
        del self.__meta_type

        d = self.__dict__.copy()
        d['__meta_type'] = t

        # restore metadata
        self.__meta_type = t

        return d

    def __setstate__(self, d):
        self.__meta_type = d['__meta_type']
        del d['__meta_type']

        self.__dict__.update(d)




import datetime
import time
from struct import pack

from types import *
from mustaine.protocol import *

# Implementation of Hessian 1.0.2 serialization
#   see: http://hessian.caucho.com/doc/hessian-1.0-spec.xtp

ENCODERS = {}
def encoder_for(data_type):
    def register(f):
        # register function `f` to encode type `data_type`
        ENCODERS[data_type] = f
        return f
    return register

def returns(data_type):
    def wrap(f):
        # wrap function `f` to return a tuple of (type,data)
        def wrapped(*args):
            return data_type, f(*args)
        return wrapped
    return wrap

def encode_object(obj):
    if type(obj) in ENCODERS:
        encoder = ENCODERS[type(obj)]
    else:
        raise TypeError("mustaine.encoder cannot serialize %s" % (type(obj),))

    return encoder(obj)[1]


@encoder_for(NoneType)
@returns('null')
def encode_null(_):
    return 'N'

@encoder_for(BooleanType)
@returns('bool')
def encode_boolean(value):
    if value:
        return 'T'
    else:
        return 'F'

@encoder_for(IntType)
@returns('int')
def encode_int(value):
    return pack('>cl', 'I', value)

@encoder_for(LongType)
@returns('long')
def encode_long(value):
    return pack('>cq', 'L', value)

@encoder_for(FloatType)
@returns('double')
def encode_double(value):
    return pack('>cd', 'D', value)

@encoder_for(datetime.datetime)
@returns('date')
def encode_date(value):
    return pack('>cq', 'd', int(time.mktime(value.timetuple())) * 1000)

@encoder_for(StringType)
@returns('string')
def encode_string(value):
    encoded = ''

    try:
        value = value.encode('ascii')
    except UnicodeDecodeError:
        raise TypeError("mustaine.encoder cowardly refuses to guess the encoding for "
                        "string objects containing bytes out of range 0x00-0x79; use "
                        "Binary or unicode objects instead")

    while len(value) > 65535:
        encoded += pack('>cH', 's', 65535)
        encoded += value[:65535]
        value    = value[65535:]

    encoded += pack('>cH', 'S', len(value.decode('utf-8')))
    encoded += value
    return encoded

@encoder_for(UnicodeType)
@returns('string')
def encode_unicode(value):
    encoded = ''

    while len(value) > 65535:
        encoded += pack('>cH', 's', 65535)
        encoded += value[:65535].encode('utf-8')
        value    = value[65535:]

    encoded += pack('>cH', 'S', len(value))
    encoded += value.encode('utf-8')
    return encoded

@encoder_for(ListType)
@returns('list')
def encode_list(obj):
    encoded = ''.join(map(encode_object, obj))
    return pack('>2cl', 'V', 'l', -1) + encoded + 'z'

@encoder_for(TupleType)
@returns('list')
def encode_tuple(obj):
    encoded = ''.join(map(encode_object, obj))
    return pack('>2cl', 'V', 'l', len(obj)) + encoded + 'z'

def encode_keyval(pair):
    return ''.join((encode_object(pair[0]), encode_object(pair[1])))

@encoder_for(DictType)
@returns('map')
def encode_map(obj):
    encoded = ''.join(map(encode_keyval, obj.items()))
    return pack('>c', 'M') + encoded + 'z'

@encoder_for(Object)
@returns('map')
def encode_mobject(obj):
    encoded  = pack('>cH', 't', len(obj._meta_type)) + obj._meta_type
    members  = obj.__getstate__()
    del members['__meta_type'] # this is here for pickling. we don't want or need it

    encoded += ''.join(map(encode_keyval, members.items()))
    return pack('>c', 'M') + encoded + 'z'

@encoder_for(Remote)
@returns('remote')
def encode_remote(obj):
    encoded = encode_string(obj.url)
    return pack('>2cH', 'r', 't', len(obj.type_name)) + obj.type_name + encoded

@encoder_for(Binary)
@returns('binary')
def encode_binary(obj):
    encoded = ''
    value   = obj.value

    while len(value) > 65535:
        encoded += pack('>cH', 'b', 65535)
        encoded += value[:65535]
        value    = value[65535:]

    encoded += pack('>cH', 'B', len(value))
    encoded += value

    return encoded

@encoder_for(Call)
@returns('call')
def encode_call(call):
    method    = call.method
    headers   = ''
    arguments = ''

    for header,value in call.headers.items():
        if not isinstance(header, StringType):
            raise TypeError("Call header keys must be strings")

        headers += pack('>cH', 'H', len(header)) + header
        headers += encode_object(value)

    # TODO: this is mostly duplicated at the top of the file in encode_object. dedup
    for arg in call.args:
        if type(arg) in ENCODERS:
            encoder = ENCODERS[type(arg)]
        else:
            raise TypeError("mustaine.encoder cannot serialize %s" % (type(arg),))

        data_type, arg = encoder(arg)
        if call.overload:
            method    += '_' + data_type
        arguments += arg

    encoded  = pack('>cBB', 'c', 1, 0)
    encoded += headers
    encoded += pack('>cH', 'm', len(method)) + method
    encoded += arguments
    encoded += 'z'

    return encoded




import datetime
from struct import unpack

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from mustaine.protocol import *

# Implementation of Hessian 1.0.2 deserialization
#   see: http://hessian.caucho.com/doc/hessian-1.0-spec.xtp

class ParseError(Exception):
    pass

class Parser(object):
    def parse_string(self, string):
        if isinstance(string, unicode):
            stream = StringIO(string.encode('utf-8'))
        else:
            stream = StringIO(string)

        return self.parse_stream(stream)

    def parse_stream(self, stream):
        self._refs   = []
        self._result = None

        if hasattr(stream, 'read') and hasattr(stream.read, '__call__'):
            self._stream = stream
        else:
            raise TypeError('Stream parser can only handle objects supporting read()')

        while True:
            code = self._read(1)

            if   code == 'c':
                if self._result:
                    raise ParseError('Encountered duplicate type header')

                version = self._read(2)
                if version != '\x01\x00':
                    raise ParseError("Encountered unrecognized call version %r" % (version,))

                self._result = Call()
                continue

            elif code == 'r':
                if self._result:
                    raise ParseError('Encountered duplicate type header')

                version = self._read(2)
                if version != '\x01\x00':
                    raise ParseError("Encountered unrecognized reply version %r" % (version,))

                self._result = Reply()
                continue

            else:
                if not self._result:
                    raise ParseError("Invalid Hessian message marker: %r" % (code,))

                if   code == 'H':
                    key, value = self._read_keyval()
                    self._result.headers[key] = value
                    continue

                elif code == 'm':
                    if not isinstance(self._result, Call):
                        raise ParseError('Encountered illegal method name within reply')

                    if self._result.method:
                        raise ParseError('Encountered duplicate method name definition')

                    self._result.method = self._read(unpack('>H', self._read(2))[0])
                    continue

                elif code == 'f':
                    if not isinstance(self._result, Reply):
                        raise ParseError('Encountered illegal fault within call')

                    if self._result.value:
                        raise ParseError('Encountered illegal extra object within reply')

                    self._result.value = self._read_fault()
                    continue

                elif code == 'z':
                    break

                else:
                    if isinstance(self._result, Call):
                        self._result.args.append(self._read_object(code))
                    else:
                        if self._result.value:
                            raise ParseError('Encountered illegal extra object within reply')

                        self._result.value = self._read_object(code)

        # have to hit a 'z' to land here, TODO derefs?
        return self._result


    def _read(self, n):
        try:
            r = self._stream.read(n)
        except IOError:
            raise ParseError('Encountered unexpected end of stream')
        except:
            raise
        else:
            if len(r) == 0:
                raise ParseError('Encountered unexpected end of stream')

        return r

    def _read_object(self, code):
        if   code == 'N':
            return None
        elif code == 'T':
            return True
        elif code == 'F':
            return False
        elif code == 'I':
            return int(unpack('>l', self._read(4))[0])
        elif code == 'L':
            return long(unpack('>q', self._read(8))[0])
        elif code == 'D':
            return float(unpack('>d', self._read(8))[0])
        elif code == 'd':
            return self._read_date()
        elif code == 's' or code == 'x':
            fragment = self._read_string()
            next     = self._read(1)
            if next.lower() == code:
                return fragment + self._read_object(next)
            else:
                raise ParseError("Expected terminal string segment, got %r" % (next,))
        elif code == 'S' or code == 'X':
            return self._read_string()
        elif code == 'b':
            fragment = self._read_binary()
            next     = self._read(1)
            if next.lower() == code:
                return fragment + self._read_object(next)
            else:
                raise ParseError("Expected terminal binary segment, got %r" % (next,))
        elif code == 'B':
            return self._read_binary()
        elif code == 'r':
            return self._read_remote()
        elif code == 'R':
            return self._refs[unpack(">L", self._read(4))[0]]
        elif code == 'V':
            return self._read_list()
        elif code == 'M':
            return self._read_map()
        else:
            raise ParseError("Unknown type marker %r" % (code,))

    def _read_date(self):
        timestamp = unpack('>q', self._read(8))[0]
        return datetime.datetime.fromtimestamp(timestamp / 1000)

    def _read_string(self):
        len = unpack('>H', self._read(2))[0]

        bytes = []
        while len > 0:
            byte = self._read(1)
            if ord(byte) in range(0x00, 0x7F):
                bytes.append(byte)
            elif ord(byte) in range(0xC2, 0xDF):
                bytes.append(byte + self._read(1))
            elif ord(byte) in range(0xE0, 0xEF):
                bytes.append(byte + self._read(2))
            elif ord(byte) in range(0xF0, 0xF4):
                bytes.append(byte + self._read(3))
            len -= 1

        return ''.join(bytes).decode('utf-8')

    def _read_binary(self):
        len = unpack('>H', self._read(2))[0]
        return Binary(self._read(len))

    def _read_remote(self):
        r    = Remote()
        code = self._read(1)

        if code == 't':
            r.type = self._read(unpack('>H', self._read(2))[0])
            code   = self._read(1)
        else:
            r.type = None

        if code != 's' and code != 'S':
            raise ParseError("Expected string object while parsing Remote object URL")

        r.url = self._read_object(code)
        return r

    def _read_list(self):
        code = self._read(1)

        if code == 't':
            # read and discard list type
            self._read(unpack('>H', self._read(2))[0])
            code = self._read(1)

        if code == 'l':
            # read and discard list length
            self._read(4)
            code = self._read(1)

        result = []
        self._refs.append(result)

        while code != 'z':
            result.append(self._read_object(code))
            code = self._read(1)

        return result

    def _read_map(self):
        code = self._read(1)

        if code == 't':
            type_len = unpack('>H', self._read(2))[0]
            if type_len > 0:
                # a typed map deserializes to an object
                result = Object(self._read(type_len))
            else:
                result = {}

            code = self._read(1)
        else:
            # untyped maps deserialize to a dict
            result = {}

        self._refs.append(result)

        fields = {}
        while code != 'z':
            key, value  = self._read_keyval(code)

            if isinstance(result, Object):
                fields[str(key)] = value
            else:
                fields[key] = value

            code = self._read(1)

        if isinstance(result, Object):
            fields['__meta_type'] = result._meta_type
            result.__setstate__(fields)
        else:
            result.update(fields)

        return result

    def _read_fault(self):
        fault = self._read_map()
        return Fault(fault['code'], fault['message'], fault.get('detail'))

    def _read_keyval(self, first=None):
        key   = self._read_object(first or self._read(1))
        value = self._read_object(self._read(1))

        return key, value




# Copyright (c) 2010, Brandon Gilmore
# Copyright (c) 2010, Phill Tornroth
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

version_info = (0, 1, 4)
__version__ =  ".".join(map(str, version_info))




import os, sys
from setuptools import find_packages, setup
from mustaine import __version__

if sys.version_info < (2,6):
    raise NotImplementedError("mustaine requires Python 2.6 or later")

setup(
    name = "mustaine",
    version = __version__,
    description = "Hessian RPC Library",
    long_description = file(
        os.path.join(
            os.path.dirname(__file__),
            'README.rst'
        )
    ).read(),

    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Object Brokering',
        'Topic :: Software Development :: Libraries',
    ],

    url = "http://github.com/bgilmore/mustaine",

    author = "Brandon Gilmore",
    author_email = "brandon@mg2.org",
    license = "BSD",

    platforms = "any",
    packages = find_packages(exclude=["test"]),
    zip_safe = True,
)

